@@ -178,7 +178,7 @@ bool KArchive::createDevice(QIODevice::OpenMode mode)
         if (!d->fileName.isEmpty()) {
             // The use of QSaveFile can't be done in the ctor (no mode known yet)
             // qCDebug(KArchiveLog) << "Writing to a file using QSaveFile";
-            d->saveFile = new QSaveFile(d->fileName);
+            d->saveFile = std::make_unique<QSaveFile>(d->fileName);
 #ifdef Q_OS_ANDROID
             // we cannot rename on to Android content: URLs
             if (d->fileName.startsWith(QLatin1String("content://"))) {
@@ -188,11 +188,11 @@ bool KArchive::createDevice(QIODevice::OpenMode mode)
             if (!d->saveFile->open(QIODevice::WriteOnly)) {
                 setErrorString(tr("QSaveFile creation for %1 failed: %2").arg(d->fileName, d->saveFile->errorString()));
 
-                delete d->saveFile;
-                d->saveFile = nullptr;
+                d->saveFile.reset();
                 return false;
             }
-            d->dev = d->saveFile;
+            d->dev = d->saveFile.get();
+            d->deviceOwned = false;
             Q_ASSERT(d->dev);
         }
         break;
@@ -229,17 +229,15 @@ bool KArchive::close()
         }
     }
 
-    if (d->dev && d->dev != d->saveFile) {
+    if (d->dev && d->dev != d->saveFile.get()) {
         d->dev->close();
     }
 
     // if d->saveFile is not null then it is equal to d->dev.
     if (d->saveFile) {
         closeSucceeded = d->saveFile->commit();
-        delete d->saveFile;
-        d->saveFile = nullptr;
-    }
-    if (d->deviceOwned) {
+        d->saveFile.reset();
+    } else if (d->deviceOwned) {
         delete d->dev; // we created it ourselves in open()
     }
 
@@ -572,8 +570,12 @@ KArchiveDirectory *KArchivePrivate::findOrCreate(const QString &path, int recurs
 
             qCDebug(KArchiveLog) << path << " is an empty file, assuming it is actually a directory and replacing";
             KArchiveEntry *myEntry = const_cast<KArchiveEntry *>(existingEntry);
-            existingEntryParentDirectory->removeEntry(myEntry);
-            delete myEntry;
+            if (existingEntryParentDirectory->removeEntryV2(myEntry)) {
+                delete myEntry;
+            } else {
+                qCDebug(KArchiveLog) << path << " is an empty file, but failed to remove it";
+                return nullptr;
+            }
         }
     }
 
@@ -644,8 +646,7 @@ void KArchivePrivate::abortWriting()
 {
     if (saveFile) {
         saveFile->cancelWriting();
-        delete saveFile;
-        saveFile = nullptr;
+        saveFile.reset();
         dev = nullptr;
     }
 }
@@ -919,10 +920,12 @@ const KArchiveFile *KArchiveDirectory::file(const QString &name) const
     return nullptr;
 }
 
+#if KARCHIVE_BUILD_DEPRECATED_SINCE(6, 13)
 void KArchiveDirectory::addEntry(KArchiveEntry *entry)
 {
-    addEntryV2(entry);
+    (void)addEntryV2(entry);
 }
+#endif
 
 bool KArchiveDirectory::addEntryV2(KArchiveEntry *entry)
 {
@@ -935,23 +938,31 @@ bool KArchiveDirectory::addEntryV2(KArchiveEntry *entry)
     return true;
 }
 
+#if KARCHIVE_BUILD_DEPRECATED_SINCE(6, 13)
 void KArchiveDirectory::removeEntry(KArchiveEntry *entry)
+{
+    (void)removeEntryV2(entry);
+}
+#endif
+
+bool KArchiveDirectory::removeEntryV2(KArchiveEntry *entry)
 {
     if (!entry) {
-        return;
+        return false;
     }
 
     QHash<QString, KArchiveEntry *>::Iterator it = d->entries.find(entry->name());
     // nothing removed?
     if (it == d->entries.end()) {
         qCWarning(KArchiveLog) << "directory " << name() << "has no entry with name " << entry->name();
-        return;
+        return false;
     }
     if (it.value() != entry) {
         qCWarning(KArchiveLog) << "directory " << name() << "has another entry for name " << entry->name();
-        return;
+        return false;
     }
     d->entries.erase(it);
+    return true;
 }
 
 bool KArchiveDirectory::isDirectory() const
